/*
	bfsha1 - Brute forces a SHA1 hash
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#ifndef COMMON_H
#define COMMON_H

#include "architecture.h"
#include <stdio.h>
#include <assert.h>

#ifdef _WIN32
	#pragma warning(disable:4996)
#endif

#ifndef __inline__
	#if defined(_MSC_VER) || defined(__INTEL_COMPILER)
		#define __inline__  __forceinline
	#elif defined(__GNUC__)
		#define __inline__  __inline __attribute__((__always_inline__))
	#else
		#define __inline__  inline
	#endif
#endif

#define __STDC_CONSTANT_MACROS
#define __STDC_FORMAT_MACROS
#define __STDC_LIMIT_MACROS
#ifdef _WIN32
	#include "inttypes.h"
	#include <time.h>
	#include <windows.h>

	#include <string.h>
	#define STR_TO_ULL(str,endPtr,radix) _strtoui64(str, endPtr, radix)
	#define STR_CMP_ICASE(a,b) strcmpi(a, b)

	typedef LARGE_INTEGER TIMER_TYPE;
	#define TIMER_FUNC(t)             QueryPerformanceCounter(&t)
	__inline__ double TIMER_DIFF(LARGE_INTEGER s, LARGE_INTEGER e)
	{
		LARGE_INTEGER f;
		QueryPerformanceFrequency(&f);
		return ((double) (e.QuadPart - s.QuadPart)) / f.QuadPart;
	}
	#define DIRECTORY_SEPERATOR       "\\"
	#define DIRECTORY_SEPERATOR_CHAR  '\\'

	#define WIN_ALIGN_16 __declspec(align(16))
	#define ALIGN_16
	#define FSEEK64(file,off,origin)  _fseeki64(file,off,origin)
	#define GET_EXE_FULL_NAME(buf,bufSize)  GetModuleFileNameA(NULL, buf, bufSize)
#else
	#include <inttypes.h>
	#include <sys/time.h>
	#include <unistd.h>
	//#include <fcntl.h>

	#include <strings.h>
	#define STR_TO_ULL(str,endPtr,radix) strtoull(str, endPtr, radix)
	#define STR_CMP_ICASE(a,b) strcasecmp(a, b)

	typedef timeval TIMER_TYPE;
	#define TIMER_FUNC(t)             gettimeofday(&t, NULL)
	#define TIMER_DIFF(s,e)           ((e.tv_sec - s.tv_sec) + (e.tv_usec - s.tv_usec) / (double)1000000.0)
	#define DIRECTORY_SEPERATOR       "/"
	#define DIRECTORY_SEPERATOR_CHAR  '/'

	#ifndef PRIu64
		#define PRIu64 "llu"
	#endif
	#ifndef PRIx64
		#define PRIx64 "llx"
	#endif
	#ifndef UINT64_MAX
		#define UINT64_MAX 0xffffffffffffffffull
	#endif
	#ifndef UINT32_MAX
		#define UINT32_MAX 0xffffffff
	#endif

	#define WIN_ALIGN_16
	#define ALIGN_16 __attribute__((aligned(16)))
	#define FSEEK64(file,off,origin)  fseeko64(file,off,origin)
	#define GET_EXE_FULL_NAME(buf,bufSize)  readlink("/proc/self/exe", buf, bufSize);
#endif

#if defined(_MSC_VER)
	#include <intrin.h>
	__inline__ uint32_t countTrailingZeros(uint32_t val)
	{
		unsigned long out;
		_BitScanForward(&out, val);
		return (uint32_t) out;
	}
	__inline__ uint32_t countLeadingZeros(uint32_t val)
	{
		unsigned long out;
		_BitScanReverse(&out, val);
		return 31 - (uint32_t) out;
	}
	#ifdef ARC_x86_32
		__inline__ uint32_t countTrailingZeros(uint64_t val)
		{
			unsigned long out;
			if ((uint32_t) val == 0)
			{
				_BitScanForward(&out, (uint32_t) (val >> 32));
				return 32 + (uint32_t) out;
			}
			else
			_BitScanForward(&out, (uint32_t) val);
			return (uint32_t) out;
		}
		__inline__ uint32_t countLeadingZeros(uint64_t val)
		{
			unsigned long out;
			if ((uint32_t) (val >> 32) != 0)
			{
				_BitScanReverse(&out, (uint32_t) (val >> 32));
				return 63 - 32 - (uint32_t) out;
			}
			_BitScanReverse(&out, (uint32_t) val);
			return 63 - (uint32_t) out;
		}
	#else
		__inline__ uint32_t countTrailingZeros(uint64_t val)
		{
			unsigned long out;
			_BitScanForward64(&out, val);
			return (uint32_t) out;
		}
		__inline__ uint32_t countLeadingZeros(uint64_t val)
		{
			unsigned long out;
			_BitScanReverse64(&out, val);
			return 63 - (uint32_t) out;
		}
	#endif
#elif defined(__INTEL_COMPILER)
	__inline__ uint32_t countTrailingZeros(uint32_t val) { return (uint32_t) _bit_scan_forward(val); }
	__inline__ uint32_t countLeadingZeros (uint32_t val) { return (uint32_t) _bit_scan_reverse(val); }
	__inline__ uint32_t countTrailingZeros(uint64_t val)
	{
		if ((uint32_t) val == 0)
		{
			return (uint32_t) _bit_scan_forward((uint32_t) (val >> 32)) + 32;
		}
		return (uint32_t) _bit_scan_forward((uint32_t) val);
	}
	__inline__ uint32_t countLeadingZeros(uint64_t val)
	{
		if ((uint32_t) (val >> 32) != 0)
		{
			return 63 - 32 - (uint32_t) _bit_scan_reverse((uint32_t) (val >> 32));
		}
		return 63 - (uint32_t) _bit_scan_reverse((uint32_t) val);
	}
#elif defined(__GNUC__)
	__inline__ uint32_t countTrailingZeros(uint32_t val) { return (uint32_t) __builtin_ctz  (val); }
	__inline__ uint32_t countLeadingZeros (uint32_t val) { return (uint32_t) __builtin_clz  (val); }
	__inline__ uint32_t countTrailingZeros(uint64_t val) { return (uint32_t) __builtin_ctzll(val); }
	__inline__ uint32_t countLeadingZeros (uint64_t val) { return (uint32_t) __builtin_clzll(val); }
#else
	__inline__ uint32_t countTrailingZeros(uint32_t val)
	{
		uint32_t ret = 0;
		if ((val & 0xffff) == 0)
		{
			val >>= 16;
			ret = 16;
		}
		if ((val & 0xff) == 0)
		{
			val >>= 8;
			ret += 8;
		}
		if ((val & 0xf) == 0)
		{
			val >>= 4;
			ret += 4;
		}
		if ((val & 0x3) == 0)
		{
			val >>= 2;
			ret += 2;
		}
		if ((val & 0x1) == 0)
		{
			ret++;
		}
		return ret;
	}
	__inline__ uint32_t countLeadingZeros (uint32_t val)
	{
		uint32_t ret = 0;
		if ((val & 0xffff0000) == 0)
		{
			val <<= 16;
			ret = 16;
		}
		if ((val & 0xff000000) == 0)
		{
			val <<= 8;
			ret += 8;
		}
		if ((val & 0xf0000000) == 0)
		{
			val <<= 4;
			ret += 4;
		}
		if ((val & 0xc0000000) == 0)
		{
			val <<= 2;
			ret += 2;
		}
		if ((val & 0x80000000) == 0)
		{
			ret++;
		}
		return 63 - ret;
	}
	__inline__ uint32_t countTrailingZeros(uint64_t val)
	{
		uint32_t val32 = (uint32_t) val, ret = 0;
		if (val32 == 0)
		{
			val32 = (uint32_t) (val >> 32);
			ret = 32;
		}
		if ((val32 & 0xffff) == 0)
		{
			val32 >>= 16;
			ret += 16;
		}
		if ((val32 & 0xff) == 0)
		{
			val32 >>= 8;
			ret += 8;
		}
		if ((val32 & 0xf) == 0)
		{
			val32 >>= 4;
			ret += 4;
		}
		if ((val32 & 0x3) == 0)
		{
			val32 >>= 2;
			ret += 2;
		}
		if ((val32 & 0x1) == 0)
		{
			ret++;
		}
		return ret;
	}
	__inline__ uint32_t countLeadingZeros(uint64_t val)
	{
		uint32_t val32 = (uint32_t) (val >> 32), ret = 0;
		if (val32 == 0)
		{
			val32 = (uint32_t) val;
			ret = 32;
		}
		if ((val32 & 0xffff0000) == 0)
		{
			val32 <<= 16;
			ret += 16;
		}
		if ((val32 & 0xff000000) == 0)
		{
			val32 <<= 8;
			ret += 8;
		}
		if ((val32 & 0xf0000000) == 0)
		{
			val32 <<= 4;
			ret += 4;
		}
		if ((val32 & 0xc0000000) == 0)
		{
			val32 <<= 2;
			ret += 2;
		}
		if ((val32 & 0x80000000) == 0)
		{
			ret++;
		}
		return 63 - ret;
	}
#endif

#define SWAP_8_LO_BITS_BIT_ENDIAN_24(x)  ((((x) & 0x01) << 23) | (((x) & 0x02) << 22) | (((x) & 0x04) << 21) | (((x) & 0x08) << 20) | \
                                          (((x) & 0x10) << 19) | (((x) & 0x20) << 18) | (((x) & 0x40) << 17) | (((x) & 0x80) << 16))
#define SWAP_8_LO_BITS_BIT_ENDIAN_32(x)  ((((x) & 0x01) << 31) | (((x) & 0x02) << 30) | (((x) & 0x04) << 29) | (((x) & 0x08) << 28) | \
                                          (((x) & 0x10) << 27) | (((x) & 0x20) << 26) | (((x) & 0x40) << 25) | (((x) & 0x80) << 24))
#define SWAP_ENDIAN_32(x)                (((x) >> 24) | ((x) << 24) | (((x) >> 8) & 0xff00) | (((x) << 8) & 0xff0000))
#define SWAP_ENDIAN_16(x)                (((x) >> 8) | ((x) << 8))

#ifdef ARC_LITTLE_ENDIAN
	#define TO_BIG_ENDIAN_32(x)     SWAP_ENDIAN_32(x)
	#define TO_BIG_ENDIAN_16(x)     SWAP_ENDIAN_16(x)
	#define TO_LITTLE_ENDIAN_32(x)  (x)
	#define TO_LITTLE_ENDIAN_16(x)  (x)
#else
	#define TO_BIG_ENDIAN_32(x)     (x)
	#define TO_BIG_ENDIAN_16(x)     (x)
	#define TO_LITTLE_ENDIAN_32(x)  SWAP_ENDIAN_32(x)
	#define TO_LITTLE_ENDIAN_16(x)  SWAP_ENDIAN_16(x)
#endif
#define FROM_BIG_ENDIAN_32(x)     TO_BIG_ENDIAN_32(x)
#define FROM_BIG_ENDIAN_16(x)     TO_BIG_ENDIAN_16(x)
#define FROM_LITTLE_ENDIAN_32(x)  TO_LITTLE_ENDIAN_32(x)
#define FROM_LITTLE_ENDIAN_16(x)  TO_LITTLE_ENDIAN_16(x)

void reportError(const char *msg, ...);
void reportWarning(const char *msg, ...);
float reciprocalRoundDown(uint64_t num);

#endif
